@@ -1,8 +1,5 @@
 /* fhandler_proc.cc: fhandler for /proc virtual filesystem
 
-   Copyright 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
-   2013, 2014, 2015 Red Hat, Inc.
-
 This file is part of Cygwin.
 
 This software is a copyrighted work licensed under the terms of the
@@ -33,6 +30,7 @@ details. */
 #include <wctype.h>
 #include "cpuid.h"
 #include "mount.h"
+#include <math.h>
 
 #define _COMPILING_NEWLIB
 #include <dirent.h>
@@ -136,14 +134,22 @@ fhandler_proc::get_proc_fhandler (const char *path)
   if (entry)
     return entry->fhandler;
 
-  int pid = atoi (path);
+  char *e;
+  pid_t pid = strtoul (path, &e, 10);
+  if (*e != '/' && *e != '\0')
+    return FH_NADA;
   pinfo p (pid);
   /* If p->pid != pid, then pid is actually the Windows PID for an execed
      Cygwin process, and the pinfo entry is the additional entry created
      at exec time.  We don't want to enable the user to access a process
      entry by using the Win32 PID, though. */
   if (p && p->pid == pid)
-    return FH_PROCESS;
+    {
+      /* Check for entry in fd subdir */
+      if (!strncmp (++e, "fd/", 3) && e[3] != '\0')
+	return FH_PROCESSFD;
+      return FH_PROCESS;
+    }
 
   bool has_subdir = false;
   while (*path)
@@ -399,6 +405,8 @@ fhandler_proc::fill_filebuf ()
   return false;
 }
 
+extern "C" int uname_x (struct utsname *);
+
 static off_t
 format_proc_version (void *, char *&destbuf)
 {
@@ -407,7 +415,7 @@ format_proc_version (void *, char *&destbuf)
   char *bufptr = buf;
   struct utsname uts_name;
 
-  uname (&uts_name);
+  uname_x (&uts_name);
   bufptr += __small_sprintf (bufptr, "%s version %s (%s@%s) (%s) %s\n",
 			  uts_name.sysname, uts_name.release, USERNAME, HOSTNAME,
 			  GCC_VERSION, uts_name.version);
@@ -421,7 +429,7 @@ static off_t
 format_proc_loadavg (void *, char *&destbuf)
 {
   extern int get_process_state (DWORD dwProcessId);
-  unsigned running = 0;
+  unsigned int running = 0;
   winpids pids ((DWORD) 0);
 
   for (unsigned i = 0; i < pids.npids; i++)
@@ -432,9 +440,17 @@ format_proc_loadavg (void *, char *&destbuf)
 	break;
     }
 
+  double loadavg[3] = { 0.0, 0.0, 0.0 };
+  getloadavg (loadavg, 3);
+
+#define HUNDRETHS(l) (int)((l - floor(l))*100)
+
   destbuf = (char *) crealloc_abort (destbuf, 48);
   return __small_sprintf (destbuf, "%u.%02u %u.%02u %u.%02u %u/%u\n",
-				    0, 0, 0, 0, 0, 0, running, pids.npids);
+			  (int)loadavg[0], HUNDRETHS(loadavg[0]),
+			  (int)loadavg[1], HUNDRETHS(loadavg[1]),
+			  (int)loadavg[2], HUNDRETHS(loadavg[2]),
+			  running, (unsigned int)pids.npids);
 }
 
 static off_t
@@ -529,9 +545,9 @@ format_proc_stat (void *, char *&destbuf)
       for (unsigned long i = 0; i < wincap.cpu_count (); i++)
 	{
 	  kernel_time += (spt[i].KernelTime.QuadPart - spt[i].IdleTime.QuadPart)
-			 * HZ / 10000000ULL;
-	  user_time += spt[i].UserTime.QuadPart * HZ / 10000000ULL;
-	  idle_time += spt[i].IdleTime.QuadPart * HZ / 10000000ULL;
+			 * CLOCKS_PER_SEC / NS100PERSEC;
+	  user_time += spt[i].UserTime.QuadPart * CLOCKS_PER_SEC / NS100PERSEC;
+	  idle_time += spt[i].IdleTime.QuadPart * CLOCKS_PER_SEC / NS100PERSEC;
 	}
 
       eobuf += __small_sprintf (eobuf, "cpu %U %U %U %U\n",
@@ -540,9 +556,10 @@ format_proc_stat (void *, char *&destbuf)
       for (unsigned long i = 0; i < wincap.cpu_count (); i++)
 	{
 	  interrupt_count += spt[i].InterruptCount;
-	  kernel_time = (spt[i].KernelTime.QuadPart - spt[i].IdleTime.QuadPart) * HZ / 10000000ULL;
-	  user_time = spt[i].UserTime.QuadPart * HZ / 10000000ULL;
-	  idle_time = spt[i].IdleTime.QuadPart * HZ / 10000000ULL;
+	  kernel_time = (spt[i].KernelTime.QuadPart - spt[i].IdleTime.QuadPart)
+			* CLOCKS_PER_SEC / NS100PERSEC;
+	  user_time = spt[i].UserTime.QuadPart * CLOCKS_PER_SEC / NS100PERSEC;
+	  idle_time = spt[i].IdleTime.QuadPart * CLOCKS_PER_SEC / NS100PERSEC;
 	  eobuf += __small_sprintf (eobuf, "cpu%d %U %U %U %U\n", i,
 				    user_time, 0ULL, kernel_time, idle_time);
 	}
@@ -593,7 +610,6 @@ format_proc_stat (void *, char *&destbuf)
   return eobuf - buf;
 }
 
-#define add_size(p,s) ((p) = ((__typeof__(p))((PBYTE)(p)+(s))))
 #define print(x) { bufptr = stpcpy (bufptr, (x)); }
 
 static inline uint32_t
@@ -632,31 +648,8 @@ format_proc_cpuinfo (void *, char *&destbuf)
   char *buf = tp.c_get ();
   char *bufptr = buf;
 
-  DWORD lpi_size = NT_MAX_PATH;
   //WORD num_cpu_groups = 1;	/* Pre Windows 7, only one group... */
-  WORD num_cpu_per_group = 64;	/* ...and a max of 64 CPUs. */
-
-  if (wincap.has_processor_groups ())
-    {
-      PSYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX lpi =
-		(PSYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX) tp.c_get ();
-      lpi_size = NT_MAX_PATH;
-      if (!GetLogicalProcessorInformationEx (RelationAll, lpi, &lpi_size))
-	lpi = NULL;
-      else
-	{
-	  PSYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX plpi = lpi;
-	  for (DWORD size = lpi_size; size > 0;
-	       size -= plpi->Size, add_size (plpi, plpi->Size))
-	    if (plpi->Relationship == RelationGroup)
-	      {
-		//num_cpu_groups = plpi->Group.MaximumGroupCount;
-		num_cpu_per_group
-			= plpi->Group.GroupInfo[0].MaximumProcessorCount;
-		break;
-	      }
-	}
-    }
+  WORD num_cpu_per_group = __get_cpus_per_group ();
 
   cpu_num_p = wcpcpy (cpu_key, L"\\Registry\\Machine\\HARDWARE\\DESCRIPTION"
 				"\\System\\CentralProcessor\\");
@@ -758,47 +751,47 @@ format_proc_cpuinfo (void *, char *&destbuf)
 	cache_alignment = clflush * 2;
       if (is_intel)
 	{
-	  uint32_t cache_level = 0;
-	  uint32_t info, layout, sets;
+	  extern long get_cpu_cache_intel (int sysc, uint32_t maxf);
+	  long cs;
 
-	  for (int idx = 0; ; ++idx)
+	  cs = get_cpu_cache_intel (_SC_LEVEL3_CACHE_SIZE, maxf);
+	  if (cs == -1)
+	    cs = get_cpu_cache_intel (_SC_LEVEL2_CACHE_SIZE, maxf);
+	  if (cs == -1)
 	    {
-	      cpuid (&info, &layout, &sets, &unused, 0x00000004, idx);
-	      uint32_t cache_type = (info & 0x1f);
-	      if (cache_type == 0)
-		break;
-	      uint32_t cur_level = ((info >> 5) & 0x7);
-	      uint32_t ways = ((layout >> 22) & 0x3ff) + 1;
-	      uint32_t part = ((layout >> 12) & 0x3ff) + 1;
-	      uint32_t line = (layout & 0xfff) + 1;
-	      sets++;
-	      if (cur_level == cache_level)
-		cache_size += ways * part * line * sets;
-	      else if (cur_level > cache_level)
-		{
-		  cache_size = ways * part * line * sets;
-		  cache_level = cur_level;
-		}
+	      cs = get_cpu_cache_intel (_SC_LEVEL1_ICACHE_SIZE, maxf);
+	      if (cs != -1)
+		cache_size = cs;
+	      cs = get_cpu_cache_intel (_SC_LEVEL1_DCACHE_SIZE, maxf);
+	      if (cs != -1)
+		cache_size += cs;
 	    }
+	  else
+	    cache_size = cs;
 	  if (cache_size != -1)
 	    cache_size >>= 10;
 	}
-      /* L2 Cache and L2 TLB Identifiers. */
-      if (cache_size == -1 && maxe >= 0x80000006)
+      else if (is_amd)
 	{
-	  uint32_t l2;
-	  cpuid (&unused, &unused, &l2, &unused, 0x80000006);
+	  extern long get_cpu_cache_amd (int sysc, uint32_t maxe);
+	  long cs;
 
-	  cache_size = l2 >> 16;
-	}
-      /* L1 Cache and TLB Identifiers. */
-      if (cache_size == -1 && maxe >= 0x80000005)
-	{
-	  uint32_t data_cache, inst_cache;
-	  cpuid (&unused, &unused, &data_cache, &inst_cache,
-		 0x80000005);
-
-	  cache_size = (inst_cache >> 24) + (data_cache >> 24);
+	  cs = get_cpu_cache_amd (_SC_LEVEL3_CACHE_SIZE, maxe);
+	  if (cs == -1)
+	    cs = get_cpu_cache_amd (_SC_LEVEL2_CACHE_SIZE, maxe);
+	  if (cs == -1)
+	    {
+	      cs = get_cpu_cache_amd (_SC_LEVEL1_ICACHE_SIZE, maxe);
+	      if (cs != -1)
+		cache_size = cs;
+	      cs = get_cpu_cache_amd (_SC_LEVEL1_DCACHE_SIZE, maxe);
+	      if (cs != -1)
+		cache_size += cs;
+	    }
+	  else
+	    cache_size = cs;
+	  if (cache_size != -1)
+	    cache_size >>= 10;
 	}
       bufptr += __small_sprintf (bufptr, "cpu family\t: %d\n"
 					 "model\t\t: %d\n"
@@ -1338,7 +1331,6 @@ format_proc_partitions (void *, char *&destbuf)
       DWORD bytes_read;
       DWORD part_cnt = 0;
       unsigned long long size;
-      device dev;
 
       restart = FALSE;
       /* ... and check for a "Harddisk[0-9]*" entry. */
@@ -1389,10 +1381,10 @@ format_proc_partitions (void *, char *&destbuf)
 			 "IOCTL_DISK_GET_PARTITION_INFO{_EX}) %E", &upath);
 	  size = 0;
 	}
-      dev.parsedisk (drive_num, 0);
+      device dev (drive_num, 0);
       bufptr += __small_sprintf (bufptr, "%5d %5d %9U %s\n",
 				 dev.get_major (), dev.get_minor (),
-				 size >> 10, dev.name + 5);
+				 size >> 10, dev.name () + 5);
       /* Fetch drive layout info to get size of all partitions on the disk. */
       if (DeviceIoControl (devhdl, IOCTL_DISK_GET_DRIVE_LAYOUT_EX,
 			   NULL, 0, ioctl_buf, NT_MAX_PATH, &bytes_read, NULL))
@@ -1414,7 +1406,7 @@ format_proc_partitions (void *, char *&destbuf)
 		      "IOCTL_DISK_GET_DRIVE_LAYOUT{_EX}): %E", &upath);
       /* Loop over partitions. */
       if (pix || pi)
-	for (DWORD i = 0; i < part_cnt; ++i)
+	for (DWORD i = 0; i < part_cnt && i < 64; ++i)
 	  {
 	    DWORD part_num;
 
@@ -1435,11 +1427,11 @@ format_proc_partitions (void *, char *&destbuf)
 	       Just skip. */
 	    if (part_num == 0)
 	      continue;
-	    dev.parsedisk (drive_num, part_num);
+	    device dev (drive_num, part_num);
 
 	    bufptr += __small_sprintf (bufptr, "%5d %5d %9U %s",
 				       dev.get_major (), dev.get_minor (),
-				       size >> 10, dev.name + 5);
+				       size >> 10, dev.name () + 5);
 	    /* Check if the partition is mounted in Windows and, if so,
 	       print the mount point list. */
 	    __small_swprintf (fpath,
@@ -1449,7 +1441,7 @@ format_proc_partitions (void *, char *&destbuf)
 		&& GetVolumePathNamesForVolumeNameW (gpath, mp_buf,
 						     NT_MAX_PATH, &len))
 	      {
-		len = strlen (dev.name + 5);
+		len = strlen (dev.name () + 5);
 		while (len++ < 6)
 		  *bufptr++ = ' ';
 		for (PWCHAR p = mp_buf; *p; p = wcschr (p, L'\0') + 1)
